@@ -1,13 +1,13 @@
 package CGI::Session;
 
-# $Id: /mirror/cgi-session/trunk/lib/CGI/Session.pm 360 2006-06-11T10:33:37.940119Z antirice  $
+# $Id: Session.pm 430 2008-07-16 00:36:15Z markstos $
 
 use strict;
 use Carp;
 use CGI::Session::ErrorHandler;
 
 @CGI::Session::ISA      = qw( CGI::Session::ErrorHandler );
-$CGI::Session::VERSION  = '4.14';
+$CGI::Session::VERSION  = '4.35';
 $CGI::Session::NAME     = 'CGISESSID';
 $CGI::Session::IP_MATCH = 0;
 
@@ -17,49 +17,76 @@ sub STATUS_DELETED  () { 4 }        # denotes session that needs deletion
 sub STATUS_EXPIRED  () { 8 }        # denotes session that was expired.
 
 sub import {
-    my $class = shift;
-    @_ or return;
+    my ($class, @args) = @_;
 
-    for(@_) {
-        $CGI::Session::IP_MATCH = ( $_ eq '-ip_match' );
+    return unless @args;
+
+  ARG:
+    foreach my $arg (@args) {
+        if ($arg eq '-ip_match') {
+            $CGI::Session::IP_MATCH = 1;
+            last ARG;
+        }
     }
 }
 
 sub new {
-    my $class   = shift;
+    my ($class, @args) = @_;
 
-    # If called as object method as in $session->new()...
     my $self;
-    if ( ref $class ) {
-        $self   = bless { %$class }, ref($class);
-        $class  = ref($class);
+    if (ref $class) {
+        #
+        # Called as an object method as in $session->new()...
+        #
+        $self  = bless { %$class }, ref( $class );
+        $class = ref $class;
         $self->_reset_status();
+        #
+        # Object may still have public data associated with it, but we
+        # don't care about that, since we want to leave that to the
+        # client's disposal. However, if new() was requested on an
+        # expired session, we already know that '_DATA' table is
+        # empty, since it was the job of flush() to empty '_DATA'
+        # after deleting. How do we know flush() was already called on
+        # an expired session? Because load() - constructor always
+        # calls flush() on all to-be expired sessions
+        #
+    }
+    else {
+        #
+        # Called as a class method as in CGI::Session->new()
+        #
 
-        # Object may still have public data associated with it, but we don't care about that,
-        # since we want to leave that to the client's disposal. However, if new() was requested on
-        # an expired session, we already know that '_DATA' table is empty, since it was the
-        # job of flush() to empty '_DATA' after deleting. How do we know flush() was already
-        # called on an expired session? Because load() - constructor always calls flush()
-        # on all to-be expired sessions
-    } else {
-        defined($self = $class->load( @_ ))
-            or return $class->set_error( "new(): failed: " . $class->errstr );
+        # Start fresh with error reporting. Errors in past objects shouldn't affect this one. 
+        $class->set_error('');
+
+        $self = $class->load( @args );
+        if (not defined $self) {
+            return $class->set_error( "new(): failed: " . $class->errstr );
+        }
     }
 
-    # Absence of '_SESSION_ID' can only signal:
-    #   * expired session
-    #       Because load() - constructor is required to empty contents of _DATA - table
-    #   * unavailable session
-    #       Such sessions are the ones that don't exist on datastore, but requested by client
-    #   * new sessions
-    #       When no specific session is requested to be loaded
-    unless ( $self->{_DATA}->{_SESSION_ID} ) {
-        $self->{_DATA}->{_SESSION_ID} = $self->_id_generator()->generate_id($self->{_DRIVER_ARGS}, $self->{_CLAIMED_ID});
-        unless ( defined $self->{_DATA}->{_SESSION_ID} ) {
-            return $self->set_error( "Couldn't generate new SID" );
+    my $dataref = $self->{_DATA};
+    unless ($dataref->{_SESSION_ID}) {
+        #
+        # Absence of '_SESSION_ID' can only signal:
+        # * Expired session: Because load() - constructor is required to
+        #                    empty contents of _DATA - table
+        # * Unavailable session: Such sessions are the ones that don't
+        #                    exist on datastore, but are requested by client
+        # * New session: When no specific session is requested to be loaded
+        #
+        my $id = $self->_id_generator()->generate_id(
+                                                     $self->{_DRIVER_ARGS},
+                                                     $self->{_CLAIMED_ID}
+                                                     );
+        unless (defined $id) {
+            return $self->set_error( "Couldn't generate new SESSION-ID" );
         }
-        $self->{_DATA}->{_SESSION_CTIME} = $self->{_DATA}->{_SESSION_ATIME} = time();
-        $self->_set_status(STATUS_NEW);
+        $dataref->{_SESSION_ID} = $id;
+        $dataref->{_SESSION_CTIME} = $dataref->{_SESSION_ATIME} = time();
+        $dataref->{_SESSION_REMOTE_ADDR} = $ENV{REMOTE_ADDR} || "";
+        $self->_set_status( STATUS_NEW );
     }
     return $self;
 }
@@ -79,8 +106,10 @@ sub is_new          { $_[0]->_test_status( STATUS_NEW ) }
 
 sub id              { return defined($_[0]->dataref) ? $_[0]->dataref->{_SESSION_ID}    : undef }
 
+# Last Access Time
 sub atime           { return defined($_[0]->dataref) ? $_[0]->dataref->{_SESSION_ATIME} : undef }
 
+# Creation Time
 sub ctime           { return defined($_[0]->dataref) ? $_[0]->dataref->{_SESSION_CTIME} : undef }
 
 sub _driver {
@@ -229,42 +258,39 @@ sub trace {}
 sub tracemsg {}
 
 sub param {
-    my $self = shift;
+    my ($self, @args) = @_;
 
-    carp "param(): attempt to read/write deleted session" if $self->_test_status(STATUS_DELETED);
-
-    #
-    # USAGE: $s->param();
-    # DESC: returns all the **public** parameters
-    unless ( @_ ) {
-        return grep { !/^_SESSION_/ } keys %{ $self->{_DATA} };
+    if ($self->_test_status( STATUS_DELETED )) {
+        carp "param(): attempt to read/write deleted session";
     }
 
-    #
-    # USAGE: $s->param($p);
+    # USAGE: $s->param();
+    # DESC:  Returns all the /public/ parameters
+    if (@args == 0) {
+        return grep { !/^_SESSION_/ } keys %{ $self->{_DATA} };
+    }
+    # USAGE: $s->param( $p );
     # DESC: returns a specific session parameter
-    return $self->{_DATA}->{$_[0]} if @_ == 1;
+    elsif (@args == 1) {
+        return $self->{_DATA}->{ $args[0] }
+    }
 
-    my %args = (
-        -name   => undef,
-        -value  => undef,
-        @_
-    );
 
-    #
-    # USAGE: $s->param(-name=>$n, -value=>$v);
-    # DESC:  updates session data using CGI.pm's 'named parameter' syntax. Only
-    # public records can be set!
-    if ( defined( $args{'-name'} ) && defined( $args{'-value'} ) ) {
-        if ( $args{'-name'} =~ m/^_SESSION_/ ) {
+    # USAGE: $s->param( -name => $n, -value => $v );
+    # DESC:  Updates session data using CGI.pm's 'named param' syntax.
+    #        Only public records can be set!
+    my %args = @args;
+    my ($name, $value) = @args{ qw(-name -value) };
+    if (defined $name && defined $value) {
+        if ($name =~ m/^_SESSION_/) {
+
             carp "param(): attempt to write to private parameter";
             return undef;
         }
-        $self->_set_status(STATUS_MODIFIED);
-        return $self->{_DATA}->{ $args{'-name'} } = $args{'-value'};
+        $self->_set_status( STATUS_MODIFIED );
+        return $self->{_DATA}->{ $name } = $value;
     }
 
-    #
     # USAGE: $s->param(-name=>$n);
     # DESC:  access to session data (public & private) using CGI.pm's 'named parameter' syntax.
     return $self->{_DATA}->{ $args{'-name'} } if defined $args{'-name'};
@@ -272,24 +298,24 @@ sub param {
     # USAGE: $s->param($name, $value);
     # USAGE: $s->param($name1 => $value1, $name2 => $value2 [,...]);
     # DESC:  updates one or more **public** records using simple syntax
-    unless ( @_ % 2 ) {
-        my $ok;
-        for ( my $i=0; $i < @_; $i += 2 ) {
-            if ( $_[$i] =~ m/^_SESSION_/) {
+    if ((@args % 2) == 0) {
+        my $modified_cnt = 0;
+	ARG_PAIR:
+        while (my ($name, $val) = each %args) {
+            if ( $name =~ m/^_SESSION_/) {
                 carp "param(): attempt to write to private parameter";
-                next;
+                next ARG_PAIR;
             }
-            $self->{_DATA}->{ $_[$i] } = $_[$i+1];
-            $ok++;
+            $self->{_DATA}->{ $name } = $val;
+            ++$modified_cnt;
         }
         $self->_set_status(STATUS_MODIFIED);
-        return $_[1] if @_ == 2 && $ok;
-        return 1;
+        return $modified_cnt;
     }
 
-    #
-    # If we reached this far none of the expected syntax were detected. Syntax error
-    croak "param(): usage error. Invalid number";
+    # If we reached this far none of the expected syntax were
+    # detected. Syntax error
+    croak "param(): usage error. Invalid syntax";
 }
 
 
@@ -426,7 +452,7 @@ sub find {
     return 1;
 }
 
-# $Id: /mirror/cgi-session/trunk/lib/CGI/Session.pm 360 2006-06-11T10:33:37.940119Z antirice  $
+# $Id: Session.pm 430 2008-07-16 00:36:15Z markstos $
 
 =pod
 
@@ -442,30 +468,38 @@ CGI::Session - persistent session data in CGI applications
 
     $CGISESSID = $session->id();
 
-    # send proper HTTP header with cookies:
+    # Send proper HTTP header with cookies:
     print $session->header();
 
-    # storing data in the session
+    # Storing data in the session:
     $session->param('f_name', 'Sherzod');
     # or
     $session->param(-name=>'l_name', -value=>'Ruzmetov');
 
-    # retrieving data
+    # Flush the data from memory to the storage driver at least before your
+    # program finishes since auto-flushing can be unreliable.
+    # Warning: A bug in your logic whereby the DBI handle has gone
+    # out of scope before flush() is called means flush() won't work
+    # (when the session is a database session), so don't do that.
+    $session->flush();
+
+    # Retrieving data:
     my $f_name = $session->param('f_name');
     # or
     my $l_name = $session->param(-name=>'l_name');
 
-    # clearing a certain session parameter
+    # Clearing a certain session parameter:
     $session->clear(["l_name", "f_name"]);
 
-    # expire '_is_logged_in' flag after 10 idle minutes:
+    # Expire '_is_logged_in' flag after 10 idle minutes:
     $session->expire('is_logged_in', '+10m')
 
-    # expire the session itself after 1 idle hour
+    # Expire the session itself after 1 idle hour:
     $session->expire('+1h');
 
-    # delete the session for good
+    # Delete the session for good:
     $session->delete();
+    $session->flush(); # Recommended practice says use flush() after delete().
 
 =head1 DESCRIPTION
 
@@ -473,6 +507,56 @@ CGI-Session is a Perl5 library that provides an easy, reliable and modular sessi
 Persistency is a key feature for such applications as shopping carts, login/authentication routines, and application that
 need to carry data across HTTP requests. CGI::Session does that and many more.
 
+=head1 A Warning about Auto-flushing
+
+As mentioned above in the Synopsis, auto-flushing can be unreliable.
+
+Consequently, you should regard it as mandatory that sessions always need to be explicitly flushed before the
+program exits.
+
+For instance, in a C<CGI::Application>-based program, C<sub teardown()> would be the appropriate place to do this.
+
+This is all part of what might be called "Object life-cycle 'v' Program life-cycle".
+
+In the simplest case the program has one object of type CGI::Session, and that object is destroyed when the
+program exits.
+
+If, however, you wish to delete objects explicitly, then each call to C<delete()> should be followed by a call
+to C<flush()>.
+
+Warning: A bug in your logic whereby the DBI handle has gone out
+out of scope before flush() is called means flush() won't work
+(when the session is a database session), so don't do that.
+
+For more detail, see the discussion of the C<delete()> method, below.
+
+=head1 A Warning about UTF8
+
+Trying to use UTF8 in a program which uses CGI::Session has lead to problems. See RT#21981 and RT#28516.
+
+In the first case the user tried "use encoding 'utf8';" in the program, and in the second case the user tried
+"$dbh->do(qq|set names 'utf8'|);".
+
+Until this problem is understood and corrected, users are advised to avoid UTF8 in conjunction with CGI::Session.
+
+For details, see: http://rt.cpan.org/Public/Bug/Display.html?id=28516 (and ...id=21981).
+
+=head1 TRANSLATIONS
+
+This document is also available in Japanese.
+
+=over 4
+
+=item o 
+
+Translation based on 4.14: http://digit.que.ne.jp/work/index.cgi?Perldoc/ja
+
+=item o
+
+Translation based on 3.11, including Cookbook and Tutorial: http://perldoc.jp/docs/modules/CGI-Session-3.11/
+
+=back
+
 =head1 TO LEARN MORE
 
 Current manual is optimized to be used as a quick reference. To learn more both about the philosophy and CGI::Session
@@ -506,17 +590,17 @@ L<Apache::Session|Apache::Session> - another fine alternative to CGI::Session.
 
 Following is the overview of all the available methods accessible via CGI::Session object.
 
-=over 4
+=head2 new()
 
-=item new()
+=head2 new( $sid )
 
-=item new( $sid )
+=head2 new( $query )
 
-=item new( $query )
+=head2 new( $dsn, $query||$sid )
 
-=item new( $dsn, $query||$sid )
+=head2 new( $dsn, $query||$sid, \%dsn_args )
 
-=item new( $dsn, $query||$sid, \%dsn_args )
+=head2 new( $dsn, $query||$sid, \%dsn_args, \%session_params )
 
 Constructor. Returns new session object, or undef on failure. Error message is accessible through L<errstr() - class method|CGI::Session::ErrorHandler/errstr>. If called on an already initialized session will re-initialize the session based on already configured object. This is only useful after a call to L<load()|/"load">.
 
@@ -534,6 +618,14 @@ If called with two arguments first will be treated as $dsn, and second will be t
     $s = CGI::Session->new("serializer:storable;id:incr", $sid);
     # etc...
 
+Briefly, C<new()> will return an initialized session object with a valid id, whereas C<load()> may return
+an empty session object with an undefined id.
+
+Tests are provided (t/new_with_undef.t and t/load_with_undef.t) to clarify the result of calling C<new()> and C<load()>
+with undef, or with an initialized CGI object with an undefined or fake CGISESSID.
+
+You are strongly advised to run the old-fashioned 'make test TEST_FILES=t/new_with_undef.t TEST_VERBOSE=1'
+or the new-fangled 'prove -v t/new_with_undef.t', for both new*.t and load*.t, and examine the output.
 
 Following data source components are supported:
 
@@ -560,17 +652,33 @@ For example, to get CGI::Session store its data using DB_File and serialize data
 
 If called with three arguments, first two will be treated as in the previous example, and third argument will be C<\%dsn_args>, which will be passed to C<$dsn> components (namely, driver, serializer and id generators) for initialization purposes. Since all the $dsn components must initialize to some default value, this third argument should not be required for most drivers to operate properly.
 
+If called with four arguments, the first three match previous examples. The fourth argument must be a hash reference with parameters to be used by the CGI::Session object. (see \%session_params above )
+
+The following is a list of the current keys:
+
+=over
+
+=item *
+
+B<name> - Name to use for the cookie/query parameter name. This defaults to CGISESSID. This can be altered or accessed by the C<name> accessor.
+
+=back
+
 undef is acceptable as a valid placeholder to any of the above arguments, which will force default behavior.
 
-=item load()
+=head2 load()
 
-=item load($query||$sid)
+=head2 load( $query||$sid )
 
-=item load($dsn, $query||$sid)
+=head2 load( $dsn, $query||$sid )
 
-=item load($dsn, $query, \%dsn_args);
+=head2 load( $dsn, $query, \%dsn_args )
 
-Constructor. Usage is identical to L<new()|/"new">, so is the return value. Major difference is, L<new()|/"new"> can create new session if it detects expired and non-existing sessions, but C<load()> does not.
+=head2 load( $dsn, $query, \%dsn_args, \%session_params )
+
+Accepts the same arguments as new(), and also returns a new session object, or
+undef on failure.  The difference is, L<new()|/"new"> can create a new session if
+it detects expired and non-existing sessions, but C<load()> does not.
 
 C<load()> is useful to detect expired or non-existing sessions without forcing the library to create new sessions. So now you can do something like this:
 
@@ -587,15 +695,26 @@ C<load()> is useful to detect expired or non-existing sessions without forcing t
         $s = $s->new() or die $s->errstr;
     }
 
-Notice, all I<expired> sessions are empty, but not all I<empty> sessions are expired!
+Notice: All I<expired> sessions are empty, but not all I<empty> sessions are expired!
+
+Briefly, C<new()> will return an initialized session object with a valid id, whereas C<load()> may return
+an empty session object with an undefined id.
+
+Tests are provided (t/new_with_undef.t and t/load_with_undef.t) to clarify the result of calling C<new()> and C<load()>
+with undef, or with an initialized CGI object with an undefined or fake CGISESSID.
+
+You are strongly advised to run the old-fashioned 'make test TEST_FILES=t/new_with_undef.t TEST_VERBOSE=1'
+or the new-fangled 'prove -v t/new_with_undef.t', for both new*.t and load*.t, and examine the output.
 
 =cut
 
-# pass a true value as the fourth parameter if you want to skip the changing of access time 
+# pass a true value as the fourth parameter if you want to skip the changing of
+# access time This isn't documented more formally, because it only called by
+# find().
 sub load {
     my $class = shift;
     return $class->set_error( "called as instance method")    if ref $class;
-    return $class->set_error( "invalid number of arguments")  if @_ > 4;
+    return $class->set_error( "Too many arguments provided to load()")  if @_ > 5;
 
     my $self = bless {
         _DATA       => {
@@ -619,9 +738,7 @@ sub load {
         _QUERY      => undef        # query object
     }, $class;
 
-    #$self->{_DATA}->{_SESSION_CTIME} = $self->{_DATA}->{_SESSION_ATIME} = time();
-
-    my ($dsn,$query_or_sid,$dsn_args,$update_atime);
+    my ($dsn,$query_or_sid,$dsn_args,$update_atime,$params);
     # load($query||$sid)
     if ( @_ == 1 ) {
         $self->_set_query_or_sid($_[0]);
@@ -630,7 +747,25 @@ sub load {
     # load($dsn, $query||$sid)
     elsif ( @_ > 1 ) {
         ($dsn, $query_or_sid, $dsn_args,$update_atime) = @_;
-        return $class->set_error( "invalid number of arguments") unless ! defined $update_atime || $update_atime =~ /^0$/;
+
+        # Make it backwards-compatible (update_atime is an undocumented key in %$params).
+        # In fact, update_atime as a key is not used anywhere in the code as yet.
+        # This patch is part of the patch for RT#33437.
+        if ( ref $update_atime and ref $update_atime eq 'HASH' ) {
+            $params = {%$update_atime};
+            $update_atime = $params->{'update_atime'};
+
+            if ($params->{'name'}) {
+                $self->{_NAME} = $params->{'name'};
+            }
+        }
+
+        # Since $update_atime is not part of the public API
+        # we ignore any value but the one we use internally: 0.
+        if (defined $update_atime and $update_atime ne '0') {
+            return $class->set_error( "Too many arguments to load(). First extra argument was: $update_atime");
+         }
+
         if ( defined $dsn ) {      # <-- to avoid 'Uninitialized value...' warnings
             $self->{_DSN} = $self->parse_dsn($dsn);
         }
@@ -642,25 +777,12 @@ sub load {
 
     }
 
-    # setting defaults, since above arguments might be 'undef'
-    $self->{_DSN}->{driver}     ||= "file";
-    $self->{_DSN}->{serializer} ||= "default";
-    $self->{_DSN}->{id}         ||= "md5";
+    $self->_load_pluggables();
 
-    # Checking and loading driver, serializer and id-generators
-    # Is this untainting reasonable here? 
-    for ( 
-            "CGI::Session::Driver::"      . ($self->{_DSN}->{driver} =~ /(.*)/)[0],
-            "CGI::Session::Serialize::"   . ($self->{_DSN}->{serializer} =~ /(.*)/)[0],
-            "CGI::Session::ID::"          . ($self->{_DSN}->{id} =~ /(.*)/)[0],
-    ) {
-        eval "require $_";
-        if ($@ ) {
-            return $self->set_error("couldn't load $_: " . $@);
-        }
-    }
+    # Did load_pluggable fail? If so, return undef, just like $class->set_error() would
+    return undef if $class->errstr;
 
-    if (not $self->{_CLAIMED_ID} ) {
+    if (not defined $self->{_CLAIMED_ID}) {
         my $query = $self->query();
         eval {
             $self->{_CLAIMED_ID} = $query->cookie( $self->name ) || $query->param( $self->name );
@@ -743,22 +865,64 @@ sub _set_query_or_sid {
 }
 
 
+sub _load_pluggables {
+    my ($self) = @_;
+
+    my %DEFAULT_FOR = (
+                       driver     => "file",
+                       serializer => "default",
+                       id         => "md5",
+                       );
+    my %SUBDIR_FOR  = (
+                       driver     => "Driver",
+                       serializer => "Serialize",
+                       id         => "ID",
+                       );
+    my $dsn = $self->{_DSN};
+    foreach my $plug qw(driver serializer id) {
+        my $mod_name = $dsn->{ $plug };
+        if (not defined $mod_name) {
+            $mod_name = $DEFAULT_FOR{ $plug };
+        }
+        if ($mod_name =~ /^(\w+)$/) {
+
+            # Looks good.  Put it into the dsn hash
+            $dsn->{ $plug } = $mod_name = $1;
+
+            # Put together the actual module name to load
+            my $prefix = join '::', (__PACKAGE__, $SUBDIR_FOR{ $plug }, q{});
+            $mod_name = $prefix . $mod_name;
+
+            ## See if we can load load it
+            eval "require $mod_name";
+            if ($@) {
+                my $msg = $@;
+                return $self->set_error("couldn't load $mod_name: " . $msg);
+            }
+        }
+        else {
+            # do something here about bad name for a pluggable
+        }
+    }
+    return;
+}
+
 =pod
 
-=item id()
+=head2 id()
 
 Returns effective ID for a session. Since effective ID and claimed ID can differ, valid session id should always
 be retrieved using this method.
 
-=item param($name)
+=head2 param($name)
 
-=item param(-name=E<gt>$name)
+=head2 param(-name=E<gt>$name)
 
 Used in either of the above syntax returns a session parameter set to $name or undef if it doesn't exist. If it's called on a deleted method param() will issue a warning but return value is not defined.
 
-=item param($name, $value)
+=head2 param($name, $value)
 
-=item param(-name=E<gt>$name, -value=E<gt>$value)
+=head2 param(-name=E<gt>$name, -value=E<gt>$value)
 
 Used in either of the above syntax assigns a new value to $name parameter,
 which can later be retrieved with previously introduced param() syntax. C<$value>
@@ -767,11 +931,11 @@ may be a scalar, arrayref or hashref.
 Attempts to set parameter names that start with I<_SESSION_> will trigger
 a warning and undef will be returned.
 
-=item param_hashref()
+=head2 param_hashref()
 
 B<Deprecated>. Use L<dataref()|/"dataref"> instead.
 
-=item dataref()
+=head2 dataref()
 
 Returns reference to session's data table:
 
@@ -782,34 +946,34 @@ Returns reference to session's data table:
 
 Useful for having all session data in a hashref, but too risky to update.
 
-=item save_param()
+=head2 save_param()
 
-=item save_param($query)
+=head2 save_param($query)
 
-=item save_param($query, \@list)
+=head2 save_param($query, \@list)
 
 Saves query parameters to session object. In other words, it's the same as calling L<param($name, $value)|/"param"> for every single query parameter returned by C<< $query->param() >>. The first argument, if present, should be either CGI object or any object which can provide param() method. If it's undef, defaults to the return value of L<query()|/"query">, which returns C<< CGI->new >>. If second argument is present and is a reference to an array, only those query parameters found in the array will be stored in the session. undef is a valid placeholder for any argument to force default behavior.
 
-=item load_param()
+=head2 load_param()
 
-=item load_param($query)
+=head2 load_param($query)
 
-=item load_param($query, \@list)
+=head2 load_param($query, \@list)
 
 Loads session parameters into a query object. The first argument, if present, should be query object, or any other object which can provide param() method. If second argument is present and is a reference to an array, only parameters found in that array will be loaded to the query object.
 
-=item clear()
+=head2 clear()
 
-=item clear('field')
+=head2 clear('field')
 
-=item clear(\@list)
+=head2 clear(\@list)
 
 Clears parameters from the session object.
 
 With no parameters, all fields are cleared. If passed a single parameter or a
 reference to an array, only the named parameters are cleared.
 
-=item flush()
+=head2 flush()
 
 Synchronizes data in memory  with the copy serialized by the driver. Call flush() 
 if you need to access the session from outside the current session object. You should
@@ -824,20 +988,27 @@ that worked with 3.x. For further details see:
  http://rt.cpan.org/Ticket/Display.html?id=17541
  http://rt.cpan.org/Ticket/Display.html?id=17299
 
-=item atime()
+Consequently, always explicitly calling C<flush()> on the session before the program exits
+should be regarded as mandatory until this problem is rectified.
+
+Warning: A bug in your logic whereby the DBI handle has gone out
+out of scope before flush() is called means flush() won't work
+(when the session is a database session), so don't do that.
+
+=head2 atime()
 
 Read-only method. Returns the last access time of the session in seconds from epoch. This time is used internally while
 auto-expiring sessions and/or session parameters.
 
-=item ctime()
+=head2 ctime()
 
 Read-only method. Returns the time when the session was first created in seconds from epoch.
 
-=item expire()
+=head2 expire()
 
-=item expire($time)
+=head2 expire($time)
 
-=item expire($param, $time)
+=head2 expire($param, $time)
 
 Sets expiration interval relative to L<atime()|/"atime">.
 
@@ -845,7 +1016,10 @@ If used with no arguments, returns the expiration interval if it was ever set. I
 
 Second form sets an expiration time. This value is checked when previously stored session is asked to be retrieved, and if its expiration interval has passed, it will be expunged from the disk immediately. Passing 0 cancels expiration.
 
-By using the third syntax you can set the expiration interval for a particular session parameter, say I<~logged-in>. This would cause the library call clear() on the parameter when its time is up. Passing 0 cancels expiration.
+By using the third syntax you can set the expiration interval for a particular
+session parameter, say I<~logged-in>. This would cause the library call clear()
+on the parameter when its time is up. Note it only makes sense to set this value to 
+something I<earlier> than when the whole session expires.  Passing 0 cancels expiration.
 
 All the time values should be given in the form of seconds. Following keywords are also supported for your convenience:
 
@@ -948,11 +1122,11 @@ sub _str2seconds {
 
 =pod
 
-=item is_new()
+=head2 is_new()
 
 Returns true only for a brand new session.
 
-=item is_expired()
+=head2 is_expired()
 
 Tests whether session initialized using L<load()|/"load"> is to be expired. This method works only on sessions initialized with load():
 
@@ -965,7 +1139,7 @@ Tests whether session initialized using L<load()|/"load"> is to be expired. This
     }
 
 
-=item is_empty()
+=head2 is_empty()
 
 Returns true for sessions that are empty. It's preferred way of testing whether requested session was loaded successfully or not:
 
@@ -980,15 +1154,24 @@ Actually, the above code is nothing but waste. The same effect could've been ach
 
 L<is_empty()|/"is_empty"> is useful only if you wanted to catch requests for expired sessions, and create new session afterwards. See L<is_expired()|/"is_expired"> for an example.
 
-=item delete()
+=head2 delete()
 
-Deletes a session from the data store and empties session data from memory, completely, so subsequent read/write requests on the same object will fail. Technically speaking, it will only set object's status to I<STATUS_DELETED> and will trigger L<flush()|/"flush">, and flush() will do the actual removal.
+Deletes a session from the data store and empties session data from memory, completely, so subsequent read/write requests on the same object will fail. Technically speaking, though, it will only set the object's status to I<STATUS_DELETED>.
 
-=item find( \&code )
+The intention is that in due course (of the program's execution) this will trigger L<flush()|/"flush">, and flush() will do the actual removal.
 
-=item find( $dsn, \&code )
+However: Auto-flushing can be unreliable, and always explicitly calling C<flush()> on the session after C<delete()>
+should be regarded as mandatory until this problem is rectified.
 
-=item find( $dsn, \&code, \%dsn_args )
+Warning: A bug in your logic whereby the DBI handle has gone out
+out of scope before flush() is called means flush() won't work
+(when the session is a database session), so don't do that.
+
+=head2 find( \&code )
+
+=head2 find( $dsn, \&code )
+
+=head2 find( $dsn, \&code, \%dsn_args )
 
 Experimental feature. Executes \&code for every session object stored in disk, passing initialized CGI::Session object as the first argument of \&code. Useful for housekeeping purposes, such as for removing expired sessions. Following line, for instance, will remove sessions already expired, but are still in disk:
 
@@ -1001,7 +1184,7 @@ Notice, above \&code didn't have to do anything, because load(), which is called
     CGI::Session->find( \&purge );
     sub purge {
         my ($session) = @_;
-        next if $session->empty;    # <-- already expired?!
+        next if $session->is_empty;    # <-- already expired?!
         if ( ($session->ctime + 3600*240) <= time() ) {
             $session->delete() or warn "couldn't remove " . $session->id . ": " . $session->errstr;
         }
@@ -1078,9 +1261,7 @@ B<Note:> find() is meant to be convenient, not necessarily efficient. It's best
 
 =head1 MISCELLANEOUS METHODS
 
-=over 4
-
-=item remote_addr()
+=head2 remote_addr()
 
 Returns the remote address of the user who created the session for the first time. Returns undef if variable REMOTE_ADDR wasn't present in the environment when the session was created.
 
@@ -1090,15 +1271,15 @@ sub remote_addr {   return $_[0]->{_DATA}->{_SESSION_REMOTE_ADDR}   }
 
 =pod
 
-=item errstr()
+=head2 errstr()
 
 Class method. Returns last error message from the library.
 
-=item dump()
+=head2 dump()
 
 Returns a dump of the session object. Useful for debugging purposes only.
 
-=item header()
+=head2 header()
 
 Replacement for L<CGI.pm|CGI>'s header() method. Without this method, you usually need to create a CGI::Cookie object and send it as part of the HTTP header:
 
@@ -1116,25 +1297,19 @@ It will retrieve the name of the session cookie from C<$session->name()> which d
 
 Now, $session->header() uses "MY_SID" as a name for the session cookie.
 
-=item query()
+=head2 query()
 
 Returns query object associated with current session object. Default query object class is L<CGI.pm|CGI>.
 
-=back
-
 =head2 DEPRECATED METHODS
 
 These methods exist solely for for compatibility with CGI::Session 3.x.
 
-=over 4
-
-=item close()
+=head3 close()
 
 Closes the session. Using flush() is recommended instead, since that's exactly what a call
 to close() does now.
 
-=back 
-
 =head1 DISTRIBUTION
 
 CGI::Session consists of several components such as L<drivers|"DRIVERS">, L<serializers|"SERIALIZERS"> and L<id generators|"ID GENERATORS">. This section lists what is available.
@@ -1190,11 +1365,6 @@ Full name: B<CGI::Session::Serialize::freezethaw>
 L<yaml|CGI::Session::Serialize::yaml> - serializes data using YAML. Requires L<YAML> or L<YAML::Syck>.
 Full name: B<CGI::Session::Serialize::yaml>
 
-=item *
-
-L<json|CGI::Session::Serialize::json> - serializes data using JSON. Requires L<JSON::Syck>.
-Full name: B<CGI::Session::Serialize::json>
-
 =back
 
 =head2 ID GENERATORS
@@ -1235,7 +1405,7 @@ CGI::Session evolved to what it is today with the help of following developers.
 
 =item Igor Plisco E<lt>igor@plisco.ruE<gt>
 
-=item Mark Stosberg E<lt>markstos@cpan.orgE<gt>
+=item Mark Stosberg 
 
 =item Matt LeBlanc E<lt>mleblanc@cpan.orgE<gt>
 
@@ -1253,11 +1423,11 @@ This library is free software. You can modify and or distribute it under the sam
 You can see what the developers have been up to since the last release by
 checking out the code repository. You can browse the Subversion repository from here:
 
- http://svn.cromedome.net/
+ http://svn.cromedome.net/repos/CGI-Session
 
 Or check it directly with C<svn> from here:
 
- svn://svn.cromedome.net/CGI-Session
+ https://svn.cromedome.net/repos/CGI-Session
 
 =head1 SUPPORT
 
@@ -1273,6 +1443,7 @@ Bug reports can be submitted at http://rt.cpan.org/NoAuth/ReportBug.html?Queue=C
 Sherzod Ruzmetov E<lt>sherzodr@cpan.orgE<gt>, http://author.handalak.com/
 
 Mark Stosberg became a co-maintainer during the development of 4.0. C<markstos@cpan.org>.
+Ron Savage became a co-maintainer during the development of 4.30. C<rsavage@cpan.org>.
 
 =head1 SEE ALSO
 
