@@ -1,6 +1,6 @@
 package CGI::Session::Driver::DBI;
 
-# $Id: /mirror/cgi-session/trunk/lib/CGI/Session/Driver/DBI.pm 340 2006-04-12T20:50:10.976508Z antirice  $
+# $Id: DBI.pm 420 2008-07-08 01:23:06Z markstos $
 
 use strict;
 
@@ -9,7 +9,7 @@ use Carp;
 use CGI::Session::Driver;
 
 @CGI::Session::Driver::DBI::ISA = ( "CGI::Session::Driver" );
-$CGI::Session::Driver::DBI::VERSION = "1.6";
+$CGI::Session::Driver::DBI::VERSION = '4.33';
 
 
 sub init {
@@ -18,6 +18,9 @@ sub init {
         if (ref $self->{Handle} eq 'CODE') {
             $self->{Handle} = $self->{Handle}->();
         }
+        else {
+            # We assume the handle is working, and there is nothing to do. 
+        }
     }
     else {
         $self->{Handle} = DBI->connect( 
@@ -44,9 +47,7 @@ sub table_name {
 
     no strict 'refs';
     if ( @_ ) {
-        my $new_name = shift;
-        $self->{TableName}           = $new_name;
-        ${ $class . "::TABLE_NAME" } = $new_name;
+        $self->{TableName} = shift;
     }
 
     unless (defined $self->{TableName}) {
@@ -64,13 +65,16 @@ sub retrieve {
 
 
     my $dbh = $self->{Handle};
-    my $sth = $dbh->prepare_cached("SELECT a_session FROM " . $self->table_name . " WHERE id=?", undef, 3);
+    my $sth = $dbh->prepare_cached("SELECT $self->{DataColName} FROM " . $self->table_name . " WHERE $self->{IdColName}=?", undef, 3);
     unless ( $sth ) {
         return $self->set_error( "retrieve(): DBI->prepare failed with error message " . $dbh->errstr );
     }
     $sth->execute( $sid ) or return $self->set_error( "retrieve(): \$sth->execute failed with error message " . $sth->errstr);
 
     my ($row) = $sth->fetchrow_array();
+
+    $sth->finish;
+
     return 0 unless $row;
     return $row;
 }
@@ -84,17 +88,20 @@ sub store {
 
 
     my $dbh = $self->{Handle};
-    my $sth = $dbh->prepare_cached("SELECT id FROM " . $self->table_name . " WHERE id=?", undef, 3);
+    my $sth = $dbh->prepare_cached("SELECT $self->{IdColName} FROM " . $self->table_name . " WHERE $self->{IdColName}=?", undef, 3);
     unless ( defined $sth ) {
         return $self->set_error( "store(): \$dbh->prepare failed with message " . $sth->errstr );
     }
 
     $sth->execute( $sid ) or return $self->set_error( "store(): \$sth->execute failed with message " . $sth->errstr );
+    my $rc = $sth->fetchrow_array;
+    $sth->finish;
+
     my $action_sth;
-    if ( $sth->fetchrow_array ) {
-        $action_sth = $dbh->prepare_cached("UPDATE " . $self->table_name . " SET a_session=? WHERE id=?", undef, 3);
+    if ( $rc ) {
+        $action_sth = $dbh->prepare_cached("UPDATE " . $self->table_name . " SET $self->{DataColName}=? WHERE $self->{IdColName}=?", undef, 3);
     } else {
-        $action_sth = $dbh->prepare_cached("INSERT INTO " . $self->table_name . " (a_session, id) VALUES(?, ?)", undef, 3);
+        $action_sth = $dbh->prepare_cached("INSERT INTO " . $self->table_name . " ($self->{DataColName}, $self->{IdColName}) VALUES(?, ?)", undef, 3);
     }
     
     unless ( defined $action_sth ) {
@@ -102,6 +109,9 @@ sub store {
     }
     $action_sth->execute($datastr, $sid)
         or return $self->set_error( "store(): \$action_sth->execute failed " . $action_sth->errstr );
+
+    $action_sth->finish;
+
     return 1;
 }
 
@@ -111,7 +121,7 @@ sub remove {
     my ($sid) = @_;
     croak "remove(): usage error" unless $sid;
 
-    my $rc = $self->{Handle}->do( 'DELETE FROM '. $self->table_name .' WHERE id= ?',{},$sid );
+   my $rc = $self->{Handle}->do( 'DELETE FROM ' . $self->table_name . " WHERE $self->{IdColName}= ?", {}, $sid );
     unless ( $rc ) {
         croak "remove(): \$dbh->do failed!";
     }
@@ -123,11 +133,16 @@ sub remove {
 sub DESTROY {
     my $self = shift;
 
+    unless ( defined $self->{Handle} && $self->{Handle} -> ping ) {
+        $self->set_error(__PACKAGE__ . '::DESTROY(). Database handle has gone away');
+        return;
+	}
+
     unless ( $self->{Handle}->{AutoCommit} ) {
-        $self->{Handle}->commit();
+        $self->{Handle}->commit;
     }
     if ( $self->{_disconnect} ) {
-        $self->{Handle}->disconnect();
+        $self->{Handle}->disconnect;
     }
 }
 
@@ -141,13 +156,16 @@ sub traverse {
     }
 
     my $tablename = $self->table_name();
-    my $sth = $self->{Handle}->prepare_cached("SELECT id FROM $tablename", undef, 3) 
+    my $sth = $self->{Handle}->prepare_cached("SELECT $self->{IdColName} FROM $tablename", undef, 3) 
         or return $self->set_error("traverse(): couldn't prepare SQL statement. " . $self->{Handle}->errstr);
     $sth->execute() or return $self->set_error("traverse(): couldn't execute statement $sth->{Statement}. " . $sth->errstr);
 
     while ( my ($sid) = $sth->fetchrow_array ) {
         $coderef->($sid);
     }
+
+    $sth->finish;
+
     return 1;
 }
 
@@ -184,10 +202,32 @@ Before you can use any DBI-based session drivers you need to make sure compatibl
 
 Your session table can define additional columns, but the above two are required. Name of the session table is expected to be I<sessions> by default. You may use a different name if you wish. To do this you have to pass I<TableName> as part of your C< \%dsn_args >:
 
-    $s = new CGI::Session("driver:sqlite", undef, {TableName=>'my_sessions'});
-    $s = new CGI::Session("driver:mysql", undef, {
-                                        TableName=>'my_sessions', 
-                                        DataSource=>'dbi:mysql:shopping_cart'});
+    $s = new CGI::Session('driver:sqlite', undef, {TableName=>'my_sessions'});
+    $s = new CGI::Session('driver:mysql', undef,
+    {
+        TableName=>'my_sessions',
+        DataSource=>'dbi:mysql:shopping_cart'.
+    });
+
+To use different column names, change the 'create table' statement, and then simply do this:
+
+    $s = new CGI::Session('driver:pg', undef,
+    {
+        TableName=>'session',
+        IdColName=>'my_id',
+        DataColName=>'my_data',
+        DataSource=>'dbi:pg:dbname=project',
+    });
+
+or
+
+    $s = new CGI::Session('driver:pg', undef,
+    {
+        TableName=>'session',
+        IdColName=>'my_id',
+        DataColName=>'my_data',
+        Handle=>$dbh,
+    });
 
 =head1 DRIVER ARGUMENTS
 
@@ -197,20 +237,27 @@ Following driver arguments are supported:
 
 =item DataSource
 
-First argument to be passed to L<DBI|DBI>->L<connect()|DBI/connect()>.
+First argument to be passed to L<DBI|DBI>->L<connect()|DBI/connect()>. If the driver makes
+the database connection itself, it will also explicitly disconnect from the database when 
+the driver object is DESTROYed.
 
 =item User
 
-User privileged to connect to the database defined in I<DataSource>.
+User privileged to connect to the database defined in C<DataSource>.
 
 =item Password
 
-Password of the I<User> privileged to connect to the database defined in I<DataSource>
+Password of the I<User> privileged to connect to the database defined in C<DataSource>
 
 =item Handle
 
-To set existing database handle object ($dbh) returned by DBI->connect(). I<Handle> will override all the
-above arguments, if any present.
+An existing L<DBI> database handle object. The handle can be created on demand
+by providing a code reference as a argument, such as C<<sub{DBI->connect}>>.
+This way, the database connection is only created if it actually needed. This can be useful
+when combined with a framework plugin like L<CGI::Application::Plugin::Session>, which creates
+a CGI::Session object on demand as well. 
+
+C<Handle> will override all the above arguments, if any present.
 
 =item TableName
 
